/*
 * Copyright (C) 2011
 * Author Roger Pau Monne <roger.pau@entel.upc.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; version 2.1 only. with the special
 * exception on linking described in file LICENSE.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 */
 
#include "libxl_osdeps.h" /* must come before any other headers */

#include "libxl_internal.h"
#include <linux/loop.h>
 
int libxl__try_phy_backend(mode_t st_mode)
{
    if (S_ISBLK(st_mode) || S_ISREG(st_mode)) {
        return 1;
    }

    return 0;
}

char *libxl__loopdev_setup(libxl__gc *gc, char *filename) {
    struct stat st;
    DIR *sysblock;
    int dir_fd;
    int control_fd = -1;
    int loop_fd = -1;
    int file_fd = -1;
    int devnr;
    struct dirent *d;
    struct loop_info64  info;
    char *loop_name = NULL;
    char *ret = NULL;

    if (stat(filename, &st) < 0) {
        LOG(ERROR, "unable to stat %s", filename);
        return NULL;
    }

    /* do not do anything if filename already is block device */
    if (S_ISBLK(st.st_mode))
        return filename;

    /* first check if file has already loop device assigned according to
     *
     * losetup implementation readdir on /sys/block + stat on
     * loop/backing_file is faster than open each loop dev
     */
    sysblock = opendir("/sys/block");
    if (!sysblock) {
        LOG(ERROR, "unable to open /sys/block");
        return NULL;
    }

    dir_fd = dirfd(sysblock);

    while ((d = readdir(sysblock))) {
        char name[300];
        struct stat buf;

        if (strcmp(d->d_name, ".") == 0
                || strcmp(d->d_name, "..") == 0
                || strncmp(d->d_name, "loop", 4) != 0)
            continue;
        snprintf(name, sizeof(name), "%s/loop/backing_file", d->d_name);
        if (fstatat(dir_fd, name, &buf, 0) == 0) {
            snprintf(name, sizeof(name), "/dev/%s", d->d_name);
            loop_fd = open(name, O_RDWR);
            if (loop_fd < 0) {
                LOG(ERROR, "unable to open %s", name);
                goto out;
            }
            if (ioctl(loop_fd, LOOP_GET_STATUS64, &info) < 0) {
                LOG(ERROR, "unable to get %s info: %s", name, strerror(errno));
                goto out;
            }
            if (info.lo_inode == st.st_ino && info.lo_device == st.st_dev) {
                /* found */
                ret = libxl__strdup(gc, name);
                goto out;
            }
            close(loop_fd);
            loop_fd = -1;
        }
    }

    /* not found, so allocate new one */
    file_fd = open(filename, O_RDWR);
    if (file_fd < 0) {
        LOG(ERROR, "unable to open %s: %s", filename, strerror(errno));
        goto out;
    }

    control_fd = open("/dev/loop-control", O_RDWR);
    if (control_fd < 0) {
        LOG(ERROR, "unable to open /dev/loop-control: %s", strerror(errno));
        goto out;
    }
    devnr = ioctl(control_fd, LOOP_CTL_GET_FREE);
    if (devnr < 0) {
        LOG(ERROR, "unable to get free loop device: %s", strerror(errno));
        goto out;
    }
    loop_name = libxl__sprintf(gc, "/dev/loop%d", devnr);
    loop_fd = open(loop_name, O_RDWR);
    if (loop_fd < 0) {
        LOG(ERROR, "unable to open %s: %s", loop_name, strerror(errno));
        goto out;
    }
    if (ioctl(loop_fd, LOOP_SET_FD, file_fd) < 0) {
        LOG(ERROR, "unable to set loop backing store: %s", strerror(errno));
        goto out;
    }
    memset(&info, 0, sizeof(info));
    strncpy((char*)info.lo_file_name, filename, LO_NAME_SIZE);
    info.lo_file_name[LO_NAME_SIZE-1] = 0;
    if (ioctl(loop_fd, LOOP_SET_STATUS64, &info) < 0) {
        LOG(ERROR, "unable to set loop info: %s", strerror(errno));
        goto out;
    }
    ret = loop_name;

out:
    if (file_fd >= 0)
        close(file_fd);
    if (control_fd >= 0)
        close(control_fd);
    if (loop_fd >= 0)
        close(loop_fd);
    if (sysblock)
        closedir(sysblock);

    return ret;
}

void libxl__loopdev_cleanup(libxl__gc *gc, char *devname)
{
    int loop_fd;

    loop_fd = open(devname, O_RDWR);
    if (loop_fd < 0) {
        LOG(ERROR, "unable to open device %s: %s", devname, strerror(errno));
        goto out;
    }

    if (ioctl(loop_fd, LOOP_CLR_FD) < 0) {
        LOG(ERROR, "unable to release device %s: %s", devname, strerror(errno));
        goto out;
    }

out:
    if (loop_fd >= 0)
        close(loop_fd);
}


char *libxl__devid_to_localdev(libxl__gc *gc, int devid)
{
    return libxl__devid_to_vdev(gc, devid);
}

/* Hotplug scripts helpers */

static char **get_hotplug_env(libxl__gc *gc,
                              char *script, libxl__device *dev)
{
    const char *type = libxl__device_kind_to_string(dev->backend_kind);
    char *be_path = libxl__device_backend_path(gc, dev);
    char **env;
    int nr = 0;

    const int arraysize = 15;
    GCNEW_ARRAY(env, arraysize);
    env[nr++] = "script";
    env[nr++] = script;
    env[nr++] = "XENBUS_TYPE";
    env[nr++] = (char *) type;
    env[nr++] = "XENBUS_PATH";
    env[nr++] = GCSPRINTF("backend/%s/%u/%d", type, dev->domid, dev->devid);
    env[nr++] = "XENBUS_BASE_PATH";
    env[nr++] = "backend";
    if (dev->backend_kind == LIBXL__DEVICE_KIND_VIF) {
        libxl_nic_type nictype;
        char *gatewaydev;

        gatewaydev = libxl__xs_read(gc, XBT_NULL,
                                    GCSPRINTF("%s/%s", be_path, "gatewaydev"));
        env[nr++] = "netdev";
        env[nr++] = gatewaydev ? : "";

        if (libxl__nic_type(gc, dev, &nictype)) {
            LOG(ERROR, "unable to get nictype");
            return NULL;
        }
        switch (nictype) {
        case LIBXL_NIC_TYPE_VIF_IOEMU:
            env[nr++] = "INTERFACE";
            env[nr++] = (char *) libxl__device_nic_devname(gc, dev->domid,
                                                    dev->devid,
                                                    LIBXL_NIC_TYPE_VIF_IOEMU);
            /*
             * We need to fall through because for PV_IOEMU nic types we need
             * to execute both the vif and the tap hotplug script, and we
             * don't know which one we are executing in this call, so provide
             * both env variables.
             */
        case LIBXL_NIC_TYPE_VIF:
            env[nr++] = "vif";
            env[nr++] = (char *) libxl__device_nic_devname(gc, dev->domid,
                                                    dev->devid,
                                                    LIBXL_NIC_TYPE_VIF);
            break;
        default:
            return NULL;
        }
    }

    env[nr++] = NULL;
    assert(nr <= arraysize);

    return env;
}

/* Hotplug scripts caller functions */

static int libxl__hotplug_nic(libxl__gc *gc, libxl__device *dev,
                               char ***args, char ***env,
                               libxl__device_action action, int num_exec)
{
    char *be_path = libxl__device_backend_path(gc, dev);
    char *script;
    int nr = 0, rc = 0;
    libxl_nic_type nictype;

    script = libxl__xs_read(gc, XBT_NULL, GCSPRINTF("%s/%s", be_path,
                                                             "script"));
    if (!script) {
        LOGE(ERROR, "unable to read script from %s", be_path);
        rc = ERROR_FAIL;
        goto out;
    }

    rc = libxl__nic_type(gc, dev, &nictype);
    if (rc) {
        LOG(ERROR, "error when fetching nic type");
        rc = ERROR_FAIL;
        goto out;
    }
    if (nictype == LIBXL_NIC_TYPE_VIF && num_exec != 0) {
        rc = 0;
        goto out;
    }

    *env = get_hotplug_env(gc, script, dev);
    if (!*env) {
        rc = ERROR_FAIL;
        goto out;
    }

    const int arraysize = 4;
    GCNEW_ARRAY(*args, arraysize);
    (*args)[nr++] = script;

    if (nictype == LIBXL_NIC_TYPE_VIF_IOEMU && num_exec) {
        (*args)[nr++] = (char *) libxl__device_action_to_string(action);
        (*args)[nr++] = "type_if=tap";
        (*args)[nr++] = NULL;
    } else {
        (*args)[nr++] = action == LIBXL__DEVICE_ACTION_ADD ? "online" :
                                                             "offline";
        (*args)[nr++] = "type_if=vif";
        (*args)[nr++] = NULL;
    }
    assert(nr == arraysize);
    rc = 1;

out:
    return rc;
}

static int libxl__hotplug_disk(libxl__gc *gc, libxl__device *dev,
                               char ***args, char ***env,
                               libxl__device_action action)
{
    char *be_path = libxl__device_backend_path(gc, dev);
    char *script;
    int nr = 0, rc = 0;

    script = libxl__xs_read(gc, XBT_NULL,
                            GCSPRINTF("%s/%s", be_path, "script"));
    if (!script) {
        LOG(INFO, "no script for %s", be_path);
        return 0;
    }

    *env = get_hotplug_env(gc, script, dev);
    if (!*env) {
        LOG(ERROR, "Failed to get hotplug environment");
        rc = ERROR_FAIL;
        goto error;
    }

    const int arraysize = 3;
    GCNEW_ARRAY(*args, arraysize);
    (*args)[nr++] = script;
    (*args)[nr++] = (char *) libxl__device_action_to_string(action);
    (*args)[nr++] = NULL;
    assert(nr == arraysize);

    LOG(DEBUG, "Args and environment ready");
    rc = 1;

error:
    return rc;
}

int libxl__get_hotplug_script_info(libxl__gc *gc, libxl__device *dev,
                                   char ***args, char ***env,
                                   libxl__device_action action,
                                   int num_exec)
{
    int rc;

    switch (dev->backend_kind) {
    case LIBXL__DEVICE_KIND_VBD:
        if (num_exec != 0) {
            LOG(DEBUG, "num_exec %d, not running hotplug scripts", num_exec);
            rc = 0;
            goto out;
        }
        rc = libxl__hotplug_disk(gc, dev, args, env, action);
        break;
    case LIBXL__DEVICE_KIND_VIF:
        /*
         * If domain has a stubdom we don't have to execute hotplug scripts
         * for emulated interfaces
         */
        if ((num_exec > 1) ||
            (libxl_get_stubdom_id(CTX, dev->domid) && num_exec)) {
            LOG(DEBUG, "num_exec %d, not running hotplug scripts", num_exec);
            rc = 0;
            goto out;
        }
        rc = libxl__hotplug_nic(gc, dev, args, env, action, num_exec);
        break;
    default:
        /* No need to execute any hotplug scripts */
        LOG(DEBUG, "backend_kind %d, no need to execute scripts", dev->backend_kind);
        rc = 0;
        break;
    }

out:
    return rc;
}

libxl_device_model_version libxl__default_device_model(libxl__gc *gc)
{
    return LIBXL_DEVICE_MODEL_VERSION_QEMU_XEN;
}

int libxl__pci_numdevs(libxl__gc *gc)
{
    DIR *dir;
    struct dirent *entry;
    int num_devs = 0;

    dir = opendir("/sys/bus/pci/devices");
    if (!dir) {
        LOGE(ERROR, "Cannot open /sys/bus/pci/devices");
        return ERROR_FAIL;
    }

    while ((entry = readdir(dir))) {
        if (entry->d_name[0] == '.')
            continue;
        num_devs++;
    }
    closedir(dir);

    return num_devs;
}

int libxl__pci_topology_init(libxl__gc *gc,
                             physdev_pci_device_t *devs,
                             int num_devs)
{

    DIR *dir;
    struct dirent *entry;
    int i, err = 0;

    dir = opendir("/sys/bus/pci/devices");
    if (!dir) {
        LOGE(ERROR, "Cannot open /sys/bus/pci/devices");
        return ERROR_FAIL;
    }

    i = 0;
    while ((entry = readdir(dir))) {
        unsigned int dom, bus, dev, func;

        if (entry->d_name[0] == '.')
            continue;

        if (i == num_devs) {
            LOG(ERROR, "Too many devices");
            err = ERROR_FAIL;
            errno = -ENOSPC;
            goto out;
        }

        if (sscanf(entry->d_name, "%x:%x:%x.%d", &dom, &bus, &dev, &func) < 4) {
            LOGE(ERROR, "Error processing /sys/bus/pci/devices");
            err = ERROR_FAIL;
            goto out;
        }

        devs[i].seg = dom;
        devs[i].bus = bus;
        devs[i].devfn = ((dev & 0x1f) << 3) | (func & 7);

        i++;
    }

 out:
    closedir(dir);

    return err;
}

/*
 * Local variables:
 * mode: C
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
